# -*- coding: utf-8 -*-
# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.4'
#       jupytext_version: 1.2.4
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# +
import pandas as pd
import pickle
import numpy as np
from pprint import pprint
from tqdm import tqdm
import networkx as nx
import matplotlib.pyplot as plt
import matplotlib
import copy
import warnings
from sklearn_hierarchical_classification.metrics import h_fbeta_score
from sklearn.model_selection import train_test_split
from sklearn.feature_extraction.text import TfidfTransformer
from sklearn.naive_bayes import MultinomialNB
from sklearn.pipeline import Pipeline
from sklearn.model_selection import cross_val_score
from sklearn.model_selection import cross_val_predict
from sklearn.metrics import confusion_matrix
from sklearn.feature_extraction.text import CountVectorizer 
from sklearn.model_selection import cross_validate
from sklearn.feature_extraction.text import TfidfVectorizer 
from sklearn import metrics
import time
from collections import Counter
from sklearn.metrics import make_scorer
from sklearn.metrics import precision_recall_fscore_support
from sklearn.metrics import f1_score
from contextlib import contextmanager
from networkx import all_pairs_shortest_path_length, relabel_nodes
from sklearn.preprocessing import MultiLabelBinarizer
from sklearn.metrics import classification_report
import itertools
from sklearn import svm
from sklearn.decomposition import TruncatedSVD
from sklearn.metrics import classification_report
from sklearn.pipeline import make_pipeline
from tqdm import tqdm_notebook 
from networkx import is_tree
from scipy.sparse import csr_matrix
from sklearn.base import BaseEstimator, ClassifierMixin, MetaEstimatorMixin, clone
from sklearn.dummy import DummyClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.utils.validation import check_array, check_consistent_length, check_is_fitted, check_X_y
from sklearn.utils.multiclass import check_classification_targets
from sklearn_hierarchical_classification.constants import CLASSIFIER, DEFAULT, METAFEATURES
from sklearn_hierarchical_classification.decorators import logger
from sklearn_hierarchical_classification.dummy import DummyProgress
from sklearn_hierarchical_classification.graph import make_flat_hierarchy, rollup_nodes
from sklearn_hierarchical_classification.validation import is_estimator, validate_parameters
from itertools import chain
from scipy.sparse import issparse, csr_matrix
from concurrent.futures import ThreadPoolExecutor as PoolExecutor
import threading
from itertools import repeat
import fasttext
import csv
from networkx import DiGraph, is_tree
import nltk 
from nltk.corpus import wordnet 
import collections
from nltk.stem import PorterStemmer
from nltk.stem import WordNetLemmatizer
 

# -

nltk.download('wordnet')

import os
dataPath = os.path.dirname(os.getcwd()) + "/data/"
display(dataPath)

dataPathSets = os.path.dirname(os.getcwd()) + "/data/training and testing data/"
display(dataPathSets)

warnings.filterwarnings("ignore") 

# # Load data

#icecat_data_full = pd.read_pickle(dataPathSets + 'icecat_data_train_full.pkl')
with open(dataPathSets + 'icecat_data_full_with_labels_duplicates_removed.pkl', 'rb') as f:
    icecat_data_full = pickle.load(f)

wdc_data_full = pd.read_pickle(dataPathSets + 'wdc_data_final_duplicates_removed.pkl')

rakuten_data_train = pd.read_pickle(dataPathSets + 'rakuten_data_train.pkl')

# +
#rakuten_data_train_more_than_5 = pd.read_pickle(dataPathSets + 'rakuten_data_train_more_than_5.pkl')
# -

rakuten_data_test = pd.read_pickle(dataPathSets + 'rakuten_data_test.pkl')

# # Load trees

with open(dataPathSets + 'tree_icecat.pkl', 'rb') as f:
    tree_icecat = pickle.load(f)

with open(dataPathSets + 'tree_wdc.pkl', 'rb') as f:
    tree_wdc = pickle.load(f)

with open(dataPathSets + 'tree_rakuten.pkl', 'rb') as f:
    tree_rakuten = pickle.load(f)


# # Split and plot data

# ## Split and plot Icecat

# +
def create_samples(data, data_columns, label_columns, train_size, test_size):
    # avoid instances of labels with only one count because it will make cv impossible
    cats_counts = data[label_columns[0]].value_counts()
    labels_less_x = []
    
    for item in cats_counts.iteritems(): 
        if item[1]*train_size < 10:
            labels_less_x.append(item[0])
            
    data_cleaned = data[~data[label_columns[0]].isin(labels_less_x)]
            
    # create reduced training and testing sets        
    training_data, testing_data, training_target, testing_target = train_test_split(data_cleaned[data_columns], data_cleaned[label_columns], train_size = train_size, test_size = test_size, random_state = 42,
                                                                            shuffle = True, stratify = data_cleaned[[label_columns[0]]])
    
        
    return training_data, testing_data, training_target, testing_target


# -

# transform path-columns to string so train_test_split can handle it
with tqdm(total=len(icecat_data_full)) as pbar:

    for index, row in icecat_data_full.iterrows():
        icecat_data_full.at[index, 'pathlist_ids'] = '>'.join(icecat_data_full.loc[index]['pathlist_ids'])
        icecat_data_full.at[index, 'pathlist_names'] = '>'.join(icecat_data_full.loc[index]['pathlist_names'])
        
        pbar.update(1)

icecat_data_full.to_pickle('icecat_data_full_with_labels.pkl')
# newest is in build training notebook and is based on this one since duplicates had to removed in the end

# + {"code_folding": []}
#with open(dataPathSets + 'icecat_data_full_with_labels.pkl', 'rb') as f:
#    icecat_data_full = pickle.load(f)

# +
# split into training and testing set 80:20 (stratify) and then again 80:20 into training and validation
# which makes it 64% training, 16% validation, 20% testing
icecat_data_train, icecat_data_test, icecat_data_train_target, icecat_data_test_target = train_test_split(icecat_data_full, icecat_data_full[['pathlist_names', 'Category.Name.Value', 'Category.CategoryID']], test_size = 0.2, random_state = 42,
                                                                            shuffle = True, stratify = icecat_data_full[['pathlist_names']])


icecat_data_train, icecat_data_validate, icecat_data_train_target, icecat_data_validate_target = train_test_split(icecat_data_train, icecat_data_train_target, test_size=0.2, random_state=42)



# +
title_length_icecat = list()

with tqdm_notebook(total=len(icecat_data_full)) as progress:
    for index, row in icecat_data_full.iterrows():
        if str(row['Title']) != 'nan' and str(row['Title']) != '' :            
            title_length_icecat.append(len(row['Title'].split()))            
        else:
            title_length_icecat.append(0)
        progress.update(1)

# +
desc_length_icecat = list()

with tqdm_notebook(total=len(icecat_data_full)) as progress:
    for index, row in icecat_data_full.iterrows():
        if str(row['Description.LongDesc']) != 'nan' and str(row['Description.LongDesc']) != '' :            
            desc_length_icecat.append(len(row['Description.LongDesc'].split()))            
        else:
            desc_length_icecat.append(0)
        progress.update(1)

# +
sumdesc_length_icecat = list()

with tqdm_notebook(total=len(icecat_data_full)) as progress:
    for index, row in icecat_data_full.iterrows():
        if str(row['SummaryDescription.LongSummaryDescription']) != 'nan' and str(row['SummaryDescription.LongSummaryDescription']) != '' :            
            sumdesc_length_icecat.append(len(row['SummaryDescription.LongSummaryDescription'].split()))            
        else:
            sumdesc_length_icecat.append(0)
        progress.update(1)

# +
# %matplotlib notebook
title_length_dict_icecat = collections.OrderedDict(sorted(Counter(title_length_icecat).items()))
desc_length_dict_icecat = collections.OrderedDict(sorted(Counter(desc_length_icecat).items()))
sumdesc_length_dict_icecat = collections.OrderedDict(sorted(Counter(sumdesc_length_icecat).items()))


plt.plot(title_length_dict_icecat.keys(), title_length_dict_icecat.values(), label='Title')
plt.xlim(-1,90)
plt.ylim(0,175000)
plt.plot(desc_length_dict_icecat.keys(), desc_length_dict_icecat.values(), label='Description')

plt.plot(sumdesc_length_dict_icecat.keys(), sumdesc_length_dict_icecat.values(), label='Summary Description')
#plt.plot(title_length_dict_wdc.keys(), [v/len(wdc_data_full) for v in title_length_dict_wdc.values()])
plt.xlabel("Number of words")
plt.ylabel("Frequency")
plt.legend(loc='upper right')
#plt.title("Icecat - Frequency of number of words in title")
plt.show()


#fig = plt.figure()
#ax1 = fig.add_subplot(111)
#ax1.set_xlabel("Number of words in title")
#ax1.set_ylabel("Frequency")
#ax1.plot(title_length_dict_icecat.keys(), title_length_dict_icecat.values())
##ax1.plot(title_length_dict_wdc.keys(), title_length_dict_wdc.values())
##ax1.scatter(x[40:],y[40:], s=10, c='r', marker="o", label='second')
##plt.legend(loc='upper left');
##plt.show()

# +
frequ_dict_icecat = collections.OrderedDict(icecat_data_full['Category.Name.Value'].value_counts().items())

fig, (ax1, ax2) = plt.subplots(1, 2)
fig.set_figwidth(10)
##fig.suptitle('Horizontally stacked subplots')
#ax1.hist(icecat_data_full['Category.Name.Value'].value_counts(), 10)
#ax1.set_xlabel("Number of instances in category")
#ax1.set_ylabel("Frequency")
#xlabels=['{:,.0f}'.format(x) + 'K' for x in ax1.get_xticks()/1000]
#ax1.set_xticklabels(xlabels)
distribution = icecat_data_full['Category.CategoryID'].value_counts()
label = distribution.index.tolist()
frequency = distribution.values.tolist() 
ax1.bar(range(len(label)), frequency)
ax1.set_xlabel("Rank")
ax1.set_ylabel("Frequency")



ax2.plot(frequ_dict_icecat.keys(), frequ_dict_icecat.values(), "o")
#ax2.plot(frequ_dict_wdc.keys(), frequ_dict_wdc.values(), "or")
ax2.set_xscale('log')
ax2.set_yscale('log')
ax2.set_xlabel("Rank (log scale)")
ax2.set_ylabel("Frequency (log scale)")


# +
## plot frequency of labels on log scale
## not sure why 179000 is not displayed though
# #%matplotlib notebook
#frequ_dict = collections.OrderedDict(icecat_data_full['Category.Name.Value'].value_counts().items())
#plt.plot(frequ_dict.keys(), frequ_dict.values(), "o")
##plt.plot(icecat_data_full['Category.Name.Value'].value_counts())
#
#
#plt.xlabel("Rank (log scale)")
#plt.ylabel("Frequency (log scale)")
##plt.title("Icecat - Distribution of categories (sortedby frequency)")
#plt.xscale('log', basex=10)
#plt.yscale('log', basey=10)
#plt.show()

# +
## plot histogram of labels on log scale
# #%matplotlib notebook
#
#data = icecat_data_full['Category.Name.Value'].value_counts()
#plt.hist(data, bins=range(min(data), max(data) + 1000, 1000))
##plt.hist(icecat_data_full['Category.Name.Value'].value_counts(), 175)
#plt.xlabel("Number of instances in category")
#plt.ylabel("Frequency")
#plt.title("Icecat - Histogram of instances per node")
##plt.xlim(250)
##plt.xscale('log')
##plt.yscale('log')
#plt.show()

# +
## plot distribution of label
# #%matplotlib notebook
#
#distribution = icecat_data_full['Category.CategoryID'].value_counts()
#label = distribution.index.tolist()
#frequency = distribution.values.tolist() 
#plt.bar(range(len(label)), frequency)
##plt.xticks([])
#plt.xlabel("Rank")
#plt.ylabel("Frequency")
##plt.xscale('log')
##plt.yscale('log')
#plt.show()
# -

# -> Very uneven distribution. Most labels have between 25 and 5000 instances. But this is natural in the product data, so it will not be balanced for training.

# ## Plot WDC

len(wdc_data_full)

# +
title_length_wdc = list()

with tqdm_notebook(total=len(wdc_data_full)) as progress:
    for index, row in wdc_data_full.iterrows():
        if str(row['title']) != 'nan' and str(row['title']) != '' :            
            title_length_wdc.append(len(row['title'].split()))            
        else:
            title_length_wdc.append(0)
        progress.update(1)

# +
desc_length_wdc = list()

with tqdm_notebook(total=len(wdc_data_full)) as progress:
    for index, row in wdc_data_full.iterrows():
        if str(row['desc']) != 'nan' and str(row['desc']) != '' :            
            desc_length_wdc.append(len(row['desc'].split()))            
        else:
            desc_length_wdc.append(0)
        progress.update(1)

# +
# %matplotlib notebook
title_length_dict_wdc = collections.OrderedDict(sorted(Counter(title_length_wdc).items()))
desc_length_dict_wdc = collections.OrderedDict(sorted(Counter(desc_length_wdc).items()))


plt.plot(list(title_length_dict_wdc.keys()), list(title_length_dict_wdc.values()), label='Title')
plt.xlim(-1,50)
plt.ylim(0,500)
plt.plot(list(desc_length_dict_wdc.keys()), list(desc_length_dict_wdc.values()), label='Description')

plt.xlabel("Number of words")
plt.ylabel("Frequency")
plt.legend(loc='upper right')
#plt.title("WDC - Frequency of number of words in title")
plt.show()


# +
frequ_dict_wdc = collections.OrderedDict(wdc_data_full['CategoryID'].value_counts().items())

fig, (ax1, ax2) = plt.subplots(1, 2)
fig.set_figwidth(10)
distribution = wdc_data_full['CategoryID'].value_counts()
label = distribution.index.tolist()
frequency = distribution.values.tolist() 
ax1.bar(range(len(label)), frequency)
ax1.set_xlabel("Rank")
ax1.set_ylabel("Frequency")



ax2.plot(frequ_dict_wdc.keys(), frequ_dict_wdc.values(), "o")
#ax2.plot(frequ_dict_wdc.keys(), frequ_dict_wdc.values(), "or")
ax2.set_xscale('log')
ax2.set_yscale('log')
ax2.set_xlabel("Rank (log scale)")
ax2.set_ylabel("Frequency (log scale)")


# +
# plot distribution of label
#
#distribution = wdc_data_full['CategoryName'].value_counts()
#label = distribution.index.tolist()
#frequency = distribution.values.tolist() 
##freq_sorted = [n for _,n in sorted(zip(label,frequency))]
##label = sorted(label)
#plt.bar(label, frequency, color="blue")
#plt.xticks([])
##plt.ylim(0, 25000)
##plt.xticks(rotation='vertical')
##plt.savefig("label_distribution_unbalanced.png")
#plt.show()
# -

# ## Split and plot Rakuten

# instances with only 1 example have to be filtered out for stratified train test split
# they are appended to training set again
only_1_rakuten = [k for k,v in rakuten_data_train['path'].value_counts().items() if rakuten_data_train['path'].value_counts()[k] < 2]

with open('only_1_rakuten.pkl', 'wb') as f:
    pickle.dump(only_1_rakuten, f)

with open(dataPath + 'only_1_rakuten.pkl', 'rb') as f:
    only_1_rakuten = pickle.load(f)

only_1_rakuten_df = rakuten_data_train[rakuten_data_train['path'].isin(only_1_rakuten)]


rakuten_train_more_than_1 = rakuten_data_train[~rakuten_data_train['path'].isin(only_1_rakuten)]

# +
# split into training and testing set 80:20 (stratify) and then again 80:20 into training and validation
# which makes it 64% training, 16% validation, 20% testing
rakuten_data_train, rakuten_data_validate, rakuten_data_train_target, rakuten_data_validate_target = train_test_split(rakuten_train_more_than_1, rakuten_train_more_than_1[['path', 'CategoryID', 'pathlist']], test_size = 0.2, random_state = 42,
                                                                            shuffle = True, stratify = rakuten_train_more_than_1[['path']])

rakuten_data_train = pd.concat([rakuten_data_train,only_1_rakuten_df])

assert len(rakuten_data_train)+len(rakuten_data_validate) == 800000

rakuten_data_full = pd.concat([rakuten_data_train,rakuten_data_test, rakuten_data_validate])
assert len(rakuten_data_full)==1000000

# +
#rakuten_data_full = pd.concat([rakuten_data_train,rakuten_data_test])
title_length_rakuten = list()

with tqdm_notebook(total=len(rakuten_data_full)) as progress:
    for index, row in rakuten_data_full.iterrows():
        if str(row['title']) != 'nan' and str(row['title']) != '' :            
            title_length_rakuten.append(len(row['title'].split()))            
        else:
            title_length_rakuten.append(0)
        progress.update(1)

# +
# %matplotlib notebook
title_length_dict_rakuten = collections.OrderedDict(sorted(Counter(title_length_rakuten).items()))


plt.plot(list(title_length_dict_rakuten.keys()), list(title_length_dict_rakuten.values()), label='Title')
#plt.xlim(-1,50)
#plt.ylim(0,500)

plt.xlabel("Number of words")
plt.ylabel("Frequency")
plt.legend(loc='upper right')
#plt.title("Rakuten - Frequency of number of words in title")
plt.show()


# +
frequ_dict_rakuten = collections.OrderedDict(rakuten_data_full['CategoryID'].value_counts().items())

fig, (ax1, ax2) = plt.subplots(1, 2)
fig.set_figwidth(10)
distribution = rakuten_data_full['CategoryID'].value_counts()
label = distribution.index.tolist()
frequency = distribution.values.tolist() 
ax1.bar(range(len(label)), frequency)
ax1.set_xlabel("Rank")
ax1.set_ylabel("Frequency")



ax2.plot(list(frequ_dict_rakuten.keys()), list(frequ_dict_rakuten.values()), "o")
#ax2.plot(frequ_dict_wdc.keys(), frequ_dict_wdc.values(), "or")
ax2.set_xscale('log')
ax2.set_yscale('log')
ax2.set_xlabel("Rank (log scale)")
ax2.set_ylabel("Frequency (log scale)")


# +
## plot distribution of label
#
#distribution = rakuten_data_train['path'].value_counts()
#label = distribution.index.tolist()
#frequency = distribution.values.tolist() 
##freq_sorted = [n for _,n in sorted(zip(label,frequency))]
##label = sorted(label)
#plt.bar(label, frequency, color="blue")
#plt.xticks([])
##plt.ylim(0, 7000)
##plt.xticks(rotation='vertical')
##plt.savefig("label_distribution_unbalanced.png")
#plt.show()

# +
#distribution

# +
## plot distribution of label
#
#distribution = rakuten_data_train['path'].value_counts()
#label = distribution.index.tolist()
#frequency = distribution.values.tolist() 
##freq_sorted = [n for _,n in sorted(zip(label,frequency))]
##label = sorted(label)
#plt.bar(label, frequency, color="blue")
#plt.xticks([])
#plt.ylim(0, 3000)
##plt.xticks(rotation='vertical')
##plt.savefig("label_distribution_unbalanced.png")
#plt.show()
# -

# # Define evaluation methods

# ## Traditional evaluation

# cv
def cross_validate_traditional(name, pipeline, X, y, all_labels=True, cv=5):
    if all_labels:
        def pr(y_true, y_pred): return precision_recall_fscore_support(y_true, y_pred, pos_label=None, average='weighted')[0]
        def re(y_true, y_pred): return precision_recall_fscore_support(y_true, y_pred, pos_label=None, average='weighted')[1]
        def f1(y_true, y_pred): return precision_recall_fscore_support(y_true, y_pred, pos_label=None, average='weighted')[2]
    else:
        def pr(y_true, y_pred): return precision_recall_fscore_support(y_true, y_pred, pos_label=None, labels=np.unique(y_pred), average='weighted')[0]
        def re(y_true, y_pred): return precision_recall_fscore_support(y_true, y_pred, pos_label=None, labels=np.unique(y_pred), average='weighted')[1]
        def f1(y_true, y_pred): return precision_recall_fscore_support(y_true, y_pred, pos_label=None, labels=np.unique(y_pred), average='weighted')[2]
    
    scoring = {'pr': make_scorer(pr), 're': make_scorer(re), 'f1': make_scorer(f1)}
    eval_measures = cross_validate(pipeline, X, y, return_train_score=False, verbose=2, n_jobs=-1, scoring=scoring, cv=cv)

    print("{} | prec_weighted: {:4f} | rec_weighted: {:4f} | f1_weighted: {:4f}".format(
    name, np.mean(eval_measures['test_pr']), np.mean(eval_measures['test_re']), np.mean(eval_measures['test_f1'])))       
    return eval_measures


# regular
def testing_traditional(y_true, y_pred, all_labels):
    if all_labels:
        return precision_recall_fscore_support(y_true, y_pred, pos_label=None, average='weighted')
    else:
        return precision_recall_fscore_support(y_true, y_pred, pos_label=None, labels=np.unique(y_true), average='weighted')


def eval_traditional(name, y_true, y_pred):
    #classifier = pipeline.fit(training_data, training_target)
    #y_pred = classifier.predict(testing_data) 
    scores_all_labels_false = testing_traditional(y_true, y_pred, all_labels=False)
    scores_all_labels_true = testing_traditional(y_true, y_pred, all_labels=True)
    
    print("{} | prec_weighted: {:4f} | rec_weighted: {:4f} | f1_weighted: {:4f} | true labels".format(
    name, scores_all_labels_false[0], scores_all_labels_false[1], scores_all_labels_false[2]))
    
    print()
    
    print("{} | prec_weighted: {:4f} | rec_weighted: {:4f} | f1_weighted: {:4f} | all labels".format(
    name, scores_all_labels_true[0], scores_all_labels_true[1], scores_all_labels_true[2]))
    
    return scores_all_labels_false, scores_all_labels_true


# ## Hierarchical evaluation

# cv
def cross_validate_hierarchical(name, pipeline, X, y, tree, root, cv=5):
    
    scoring = make_scorer(hierarchical_eval, tree=tree, root=root)
    
    eval_measures = cross_validate(pipeline, X, y, return_train_score=False, n_jobs=-1, scoring=scoring, cv=cv)

    print("{} | cross-validated h_f_score: {:4f}".format(
    name, np.mean(eval_measures['test_score'])))       
    return eval_measures


# +
# regular

@contextmanager
def multi_labeled(y_true, y_pred, graph, root):

    mlb = MultiLabelBinarizer()
    all_classes = [
        node
        for node in graph.nodes
        if node != root
    ]
    #print('all_classes',all_classes)
    # Nb. we pass a (singleton) list-within-a-list as fit() expects an iterable of iterables -> Changed implementation here
    all_classes_new=[]
    for klasse in all_classes:
        all_classes_new.append([klasse])
        
    mlb.fit(all_classes_new)
    
    y_true_new=[]
    for klasse in y_true:
        y_true_new.append([klasse])
        
    y_pred_new=[]
    for klasse in y_pred:
        y_pred_new.append([klasse])

    node_label_mapping = {
        old_label: new_label
        for new_label, old_label in enumerate(list(mlb.classes_))
    }
    #print('node_label_mapping',node_label_mapping)
    #print('y_true transform',y_true)
    #print('mlb.transform(y_true)',mlb.transform(y_true_new)[0])
    yield (
        mlb.transform(y_true_new),
        mlb.transform(y_pred_new),
        relabel_nodes(graph, node_label_mapping),
        root,
    )


def fill_ancestors(y, root, graph, copy=True):

    y_ = y.copy() if copy else y
    paths = all_pairs_shortest_path_length(graph.reverse(copy=False))
    for target, distances in paths:
        if target == root:
            # Our stub ROOT node, can skip
            continue
        #print('y',y)
        #print('target',target)
        ix_rows = np.where(y[:, target] > 0)[0]
        # all ancestors, except the last one which would be the root node
        ancestors = list(distances.keys())[:-1]
        #print('ancestors',ancestors)
        #print('mesh',tuple(np.meshgrid(ix_rows, ancestors)))
        #print('target',target)
        #print(type(target))
        y_[tuple(np.meshgrid(ix_rows, ancestors))] = 1
    graph.reverse(copy=False)
    #print('y_',y_[0])
    return y_


def h_precision_score(y_true, y_pred, class_hierarchy, root):

    y_true_ = fill_ancestors(y_true, root, graph=class_hierarchy)
    y_pred_ = fill_ancestors(y_pred, root, graph=class_hierarchy)
    
    #print('y_true_',y_true_[64])
    #print('y_pred_',y_pred_[64])

    ix = np.where((y_true_ != 0) & (y_pred_ != 0))

    true_positives = len(ix[0])
    all_results = np.count_nonzero(y_pred_)
    
    #print('all_results', all_results)
    return true_positives / all_results


def h_recall_score(y_true, y_pred, class_hierarchy, root):

    y_true_ = fill_ancestors(y_true, root, graph=class_hierarchy)
    y_pred_ = fill_ancestors(y_pred, root, graph=class_hierarchy)

    ix = np.where((y_true_ != 0) & (y_pred_ != 0))

    true_positives = len(ix[0])
    all_positives = np.count_nonzero(y_true_)
    #print('all_positives',all_positives)

    return true_positives / all_positives


def h_fbeta_score(y_true, y_pred, class_hierarchy, root, beta=1.):
    #print('HI', y_true[197])
    #print('YEE',class_hierarchy.nodes)

    hP = h_precision_score(y_true, y_pred, class_hierarchy, root)
    hR = h_recall_score(y_true, y_pred, class_hierarchy, root)
    return (1. + beta ** 2.) * hP * hR / (beta ** 2. * hP + hR)


# -

def hierarchical_eval(name, y_true, y_pred, tree, root):

    with multi_labeled(y_true, y_pred, tree, root) as (y_test_, y_pred_, graph_, root_):
        h_fbeta = h_fbeta_score(
            y_test_,
            y_pred_,
            graph_,
            root_,
        )
        #print("h_f1_score: ", h_fbeta)
        print("{} | h_f1: {:4f}".format(name, h_fbeta))
        return h_fbeta


# # Experiments

# +
# Dict-based
e_0_a = 'Dict-based; full set; title only; all features'
e_0_aa = 'Dict-based; full set; title only; all features; synonyms'
e_0_b = 'Dict-based; full set; title only; all features; lemmatizing'
e_0_bb = 'Dict-based; full set; title only; all features; lemmatizing; synonyms'
e_0_c = 'Dict-based; full set; title only; all features; synonyms; fallback classifier'

# Flat
e_1 = 'Naive Bayes; full training set; title only; all features; Tf-idf'
e_1_wdc = 'Naive Bayes; full training set; title and desc; all features; Tf-idf'
e_1_icecat_a = 'Naive Bayes; full training set; title and long desc; all features; Tf-idf'
e_1_icecat_b = 'Naive Bayes; full training set; title and long desc and summary desc; all features; Tf-idf'



# LCPN
e_2 = 'Naive Bayes; full training set; title only; all features; Tf-idf'
e_2_wdc = 'Naive Bayes; full training set; title and desc; all features; Tf-idf'

# Fasttext
e_3 = 'Fasttext; full training set; title only; all features'
e_3_wdc = 'Fasttext; full training set; title and desc; all features'
# -

# # Dictionary-based approach

# +
# get class names and synonyms, but only for leaf noded!
icecat_synonyms_dict = {}
wdc_synonyms_dict = {}
leaf_nodes_icecat = [node[0] for node in tree_icecat.out_degree if node[1]==0]
leaf_nodes_wdc = [node[0] for node in tree_wdc.out_degree if node[1]==0]


with tqdm_notebook(total=len(leaf_nodes_icecat), desc="Creating icecat_synonyms_dict") as progress:
    for classname in leaf_nodes_icecat:
        synonyms = [] 

        for syn in wordnet.synsets(classname.replace(' ', '_')): 
            for l in syn.lemmas(): 
                synonyms.append(l.name().replace('_', ' ').lower()) 
        icecat_synonyms_dict[classname] = synonyms
        progress.update(1)
    
with tqdm_notebook(total=len(leaf_nodes_wdc), desc="Creating icecat_synonyms_dict") as progress2:

    for classname in leaf_nodes_wdc:
        synonyms = [] 
  
        for syn in wordnet.synsets(classname.replace(' ', '_')): 
            for l in syn.lemmas(): 
                synonyms.append(l.name().replace('_', ' ').lower())
        wdc_synonyms_dict[classname] = synonyms
        progress2.update(1)

# +
# fallback classifier (not currently used)
pipeline = Pipeline([
('vect', CountVectorizer()),
#('tfidf', TfidfTransformer()),
('clf', MultinomialNB()),
])

classifier_dictionary_based = pipeline.fit(icecat_data_train['Title'].values, icecat_data_train_target['Category.Name.Value'].values)


# + {"code_folding": []}
wnl = WordNetLemmatizer()
porter = PorterStemmer()

def count_occurances(list_of_words, target_sring, lemmatize=False):    
    count = 0
    if lemmatize:
        for word in list_of_words:
            count = count + target_sring.count(wnl.lemmatize(word))
    else:
        for word in list_of_words:
            count = count + target_sring.count(word)
    
    # division by len(list_of_words) because if a label has multiple words, its count is likely to be higher
    return count/len(list_of_words)



def classify_dictionary_based(test_data, set_name, classifier=classifier_dictionary_based, synonyms=False, lemmatize=False, fallback_classifier=False):
    y_pred = []
    count = 0
    to_predict_by_fallback = []
 
    if set_name=='icecat':
        synonyms_dict = icecat_synonyms_dict
        most_frequent = 'Notebooks'

    else:
        synonyms_dict = wdc_synonyms_dict
        most_frequent = 'Ink Cartridges'

    
    if synonyms:
        with tqdm_notebook(total=len(test_data), desc="Getting predictions...") as progress:

            for text_of_instance in test_data:
                word_count = {}

                for classname in synonyms_dict.keys():
                    words_in_classname = set(classname.lower().split() + ' '.join(synonyms_dict[classname]).split())
                    word_count[classname] = count_occurances(words_in_classname, text_of_instance, lemmatize)

                max_words = max(word_count, key=word_count.get)
                
                if word_count[max_words] == 0:
                    count = count+1
                    if fallback_classifier:                     
                        to_predict_by_fallback.append(text_of_instance)
                        y_pred.append(0)
                    else:
                        y_pred.append(most_frequent)
                else:
                    y_pred.append(max(word_count, key=word_count.get))
                progress.update(1)
                
            print('Most frequent class/fallback classifier was used %d times' %(count))
        
    else:
        with tqdm_notebook(total=len(test_data), desc="Getting predictions...") as progress:

            for text_of_instance in test_data:
                word_count = {}

                for classname in synonyms_dict.keys():
                    words_in_classname = classname.lower().split()
                    word_count[classname] = count_occurances(words_in_classname, text_of_instance, lemmatize)

                max_words = max(word_count, key=word_count.get)
                
                if word_count[max_words] == 0:
                    count = count+1
                    if fallback_classifier:
                        to_predict_by_fallback.append(text_of_instance)
                        y_pred.append(0)
                    else:
                        y_pred.append(most_frequent)
                else:
                    y_pred.append(max(word_count, key=word_count.get))
                progress.update(1)
                
            print('Most frequent class/fallback classifier was used %d times' %(count))


    if fallback_classifier:         
        fallback_predictions = classifier.predict(to_predict_by_fallback)

        for i, p in enumerate(y_pred):
            if p==0:
                y_pred[i]=fallback_predictions[0]
                fallback_predictions = fallback_predictions[1:]
            
    return y_pred


# +
def classify_dictionary_based_parallel(row, set_name, classifier=classifier_dictionary_based, synonyms=False, lemmatize=False, fallback_classifier=False):
    #y_pred = []
    #count = 0
 
    if set_name=='icecat':
        synonyms_dict = icecat_synonyms_dict
        most_frequent = 'Notebooks'

    else:
        synonyms_dict = wdc_synonyms_dict
        most_frequent = 'Ink Cartridges'

    
    if synonyms:

            #for text_of_instance in test_data:
                word_count = {}

                for classname in synonyms_dict.keys():
                    words_in_classname = set(classname.lower().split() + ' '.join(synonyms_dict[classname]).split())
                    word_count[classname] = count_occurances(words_in_classname, row, lemmatize)

                max_words = max(word_count, key=word_count.get)
                
                if word_count[max_words] == 0:
                    #count = count+1
                    if fallback_classifier:                     
                        return classifier.predict([row])
                    else:
                        return most_frequent
                else:
                    return max(word_count, key=word_count.get)
                #progress.update(1)
                
            #print('Most frequent class was used %d times' %(count))
        
    else:

            #for text_of_instance in test_data:
                word_count = {}

                for classname in synonyms_dict.keys():
                    words_in_classname = classname.lower().split()
                    word_count[classname] = count_occurances(words_in_classname, row, lemmatize)

                max_words = max(word_count, key=word_count.get)
                
                if word_count[max_words] == 0:
                    #count = count+1
                    if fallback_classifier:
                        return classifier.predict([row])
                    else:
                        return most_frequent
                else:
                    return max(word_count, key=word_count.get)
                #progress.update(1)
                
            #print('Most frequent class was used %d times' %(count))

             
            
        #
        #words = text_of_instance.split()
        #for word in words:
        #    #print(word)
        #    if word.lower() in synonyms_dict.keys():
        #        y_pred.append(word)
        #        break
        #    else:
        #        y_pred.append(most_frequent)
        #        break
        
    
    #return y_pred
# -

# ## Dictionary-based approach Icecat

y_pred = classify_dictionary_based(icecat_data_test['Title'], 'icecat')

scores_all_labels_false, scores_all_labels_true = eval_traditional(e_0_a, icecat_data_test_target['Category.Name.Value'].values, y_pred)

y_pred = classify_dictionary_based(icecat_data_test['Title'], 'icecat', synonyms=True)

scores_all_labels_false, scores_all_labels_true = eval_traditional(e_0_aa, icecat_data_test_target['Category.Name.Value'].values, y_pred)

y_pred = classify_dictionary_based(icecat_data_test['Title'], 'icecat', lemmatize=True)

scores_all_labels_false, scores_all_labels_true = eval_traditional(e_0_b, icecat_data_test_target['Category.Name.Value'].values, y_pred)

y_pred = classify_dictionary_based(icecat_data_test['Title'], 'icecat', synonyms=True, lemmatize=True)

scores_all_labels_false, scores_all_labels_true = eval_traditional(e_0_bb, icecat_data_test_target['Category.Name.Value'].values, y_pred)

y_pred = classify_dictionary_based(icecat_data_test['Title'], 'icecat', synonyms=True, fallback_classifier=True)

scores_all_labels_false, scores_all_labels_true = eval_traditional(e_0_c, icecat_data_test_target['Category.Name.Value'].values, y_pred)

# ## Dictionary-based approach WDC

y_pred = classify_dictionary_based(wdc_data_full['title'], 'wdc')

scores_all_labels_false, scores_all_labels_true = eval_traditional(e_0_a, wdc_data_full['CategoryName'].values, y_pred)

y_pred = classify_dictionary_based(wdc_data_full['title'], 'wdc', synonyms=True)

scores_all_labels_false, scores_all_labels_true = eval_traditional(e_0_aa, wdc_data_full['CategoryName'].values, y_pred)

y_pred = classify_dictionary_based(wdc_data_full['title'], 'wdc', lemmatize=True)

scores_all_labels_false, scores_all_labels_true = scores = eval_traditional(e_0_b, wdc_data_full['CategoryName'].values, y_pred)

y_pred = classify_dictionary_based(wdc_data_full['title'], 'wdc', synonyms=True, lemmatize=True)

scores_all_labels_false, scores_all_labels_true = scores = eval_traditional(e_0_bb, wdc_data_full['CategoryName'].values, y_pred)

# -> Lemmatizing improves preditions by a lot and leads to less fallbacks to most frequent class. But computationally way less expensive (more than 4x faster) than stemming, similar performance. Synonyms give performance boost to really low score, but once score is already better because of lemmatizing, it does not improve it, but it does lead to significantly less fallback predictions of most frequent class.

y_pred = classify_dictionary_based(wdc_data_full['title'], 'wdc', fallback_classifier=True, lemmatize=True)

scores_all_labels_false, scores_all_labels_true = scores = eval_traditional(e_0_c, wdc_data_full['CategoryName'].values, y_pred)


# ## Dictionary-based approach Rakuten

# -> Not possible because labels/categories are random integers

# # Flat classification

# ## Flat classification Icecat

# Naive Bayes as competitive baseline
pipeline = Pipeline([
    ('vect', CountVectorizer()),
    ('tfidf', TfidfTransformer()),
    ('clf', MultinomialNB()),
])

# Traditional testing
# Evaluate pipeline on held out testing set. Refit on whole training set first
classifier = pipeline.fit(icecat_data_train['Title'].values, icecat_data_train_target['Category.Name.Value'].values)

y_pred = classifier.predict(icecat_data_test['Title'].values)

y_true = icecat_data_test_target['Category.Name.Value'].values

# Some labels in y_true might not appear in y_pred. This means that there is no F-score to calculate for this label, 
# and thus the F-score for this case is considered to be 0.0.
scores_all_labels_false, scores_all_labels_true = eval_traditional(e_1, y_true, y_pred)

# Hierarchical testing
h_f_score = hierarchical_eval(e_1, y_true, y_pred, tree_icecat, 'Computers & Electronics')

y_pred = classifier.predict((icecat_data_test['Title'].str.strip() + ' ' + icecat_data_test['Description.LongDesc'].str.strip()).values.astype('U'))

scores_all_labels_false, scores_all_labels_true = eval_traditional(e_1_icecat_a, y_true, y_pred)

# Hierarchical testing
h_f_score = hierarchical_eval(e_1_icecat_a, y_true, y_pred, tree_icecat, 'Computers & Electronics')

# +
y_pred = classifier.predict((icecat_data_test['Title'].str.strip() + ' ' + icecat_data_test['Description.LongDesc'].str.strip() + ' ' + icecat_data_test['SummaryDescription.LongSummaryDescription'].str.strip()).values.astype('U'))


# -

scores_all_labels_false, scores_all_labels_true = eval_traditional(e_1_icecat_b, y_true, y_pred)

# Hierarchical testing
h_f_score = hierarchical_eval(e_1_icecat_b, y_true, y_pred, tree_icecat, 'Computers & Electronics')

# -> Training on title only and then test on title + other features (long desc and summary desc) does not work well! Most likely because then there are many words that did not appear during training

# +
# todo train on title, desc, und long desc and then test again
#classifier = pipeline.fit(icecat_data_train['Title'].values, icecat_data_train_target['Category.Name.Value'].values)
classifier = pipeline.fit(((icecat_data_train['Title'].str.strip() + ' ' + icecat_data_train['Description.LongDesc'].str.strip() + ' ' + icecat_data_train['SummaryDescription.LongSummaryDescription'].str.strip()).values.astype('U')), icecat_data_train_target['Category.Name.Value'].values)


# -

y_pred = classifier.predict((icecat_data_test['Title'].str.strip() + ' ' + icecat_data_test['Description.LongDesc'].str.strip() + ' ' + icecat_data_test['SummaryDescription.LongSummaryDescription'].str.strip()).values.astype('U'))


scores_all_labels_false, scores_all_labels_true = eval_traditional(e_1_icecat_a, y_true, y_pred)





# ## Flat classification WDC

# +
# Traditional testing
# Apply classifier trained on Icecat training set
# classifier.predict transforms data according to pipeline (classifier is a pipeline object) 
# and predicts with the final estimator

y_pred = classifier.predict(wdc_data_full['title'].values)
y_true = wdc_data_full['CategoryName'].values
# -

scores_all_labels_false, scores_all_labels_true = eval_traditional(e_1, y_true, y_pred)

# Hierarchical testing
h_f_score = hierarchical_eval(e_1,y_true, y_pred, tree_wdc, 'Computers & Electronics')

y_pred = classifier.predict((wdc_data_full['title'].str.strip() + ' ' + wdc_data_full['desc'].str.strip()).values)

scores_all_labels_false, scores_all_labels_true = eval_traditional(e_1_wdc, y_true, y_pred)

h_f_score = hierarchical_eval(e_1_wdc,y_true, y_pred, tree_wdc, 'Computers & Electronics')

# ## Flat classification Rakuten

rakuten_data_train.head()

# Traditional testing
classifier = pipeline.fit(rakuten_data_train['title'].values, rakuten_data_train['path'].values)


y_pred = classifier.predict(rakuten_data_test['title'].values)

y_true = rakuten_data_test['path'].values

scores_all_labels_false, scores_all_labels_true = eval_traditional(e_1, y_true, y_pred)

# +
# modify y_pred and y_true -> only take last entry as it has to be a node to make hierarchical evaluation work. It is a 
# tree structure so it is fine to do that
y_pred_modified = []
for pred in y_pred:
    y_pred_modified.append(pred.split('>')[-1])

y_true_modified = []
for true in y_true:
    y_true_modified.append(true.split('>')[-1])

# -

# Hierarchical testing
h_f_score = hierarchical_eval(e_1, y_true_modified, y_pred_modified, tree_rakuten, 'Root')

# # LCPN

# ## LCPN classic

# +
"""Helpers for workings with sequences and (numpy) arrays."""

def flatten_list(lst):
    return list(chain(*lst))


def apply_along_rows(self, func, progress,X):
    """
    Apply function row-wise to input matrix X.
    This will work for dense matrices (eg np.ndarray)
    as well as for CSR sparse matrices.
    """
    if issparse(X):
        #with self._progress(total=self.n_classes_ + 1, desc="Predicting") as progress:
        return np.array([
            func(X.getrow(i),progress)
            for i in range(X.shape[0])
        ])
    else:
        # XXX might break vis-a-vis this issue merging: https://github.com/numpy/numpy/pull/8511
        # See discussion over issue with truncated string when using np.apply_along_axis here:
        #   https://github.com/numpy/numpy/issues/8352
        return np.ma.apply_along_axis(
            lambda x: func(x.reshape(1, -1)),
            axis=1,
            arr=X,
        )


def apply_rollup_Xy(X, y):
    """
    Parameters
    ----------
    X : (sparse) array-like, shape = [n_samples, n_features]
        Data.
    y : list-of-lists - [n_samples]
        For each sample, y maintains list of labels this sample should be used for in training.
    Returns
    -------
    X_, y_
        Transformed by 'flattening' out y parameter and duplicating corresponding rows in X
    """
    # Compute number of rows we will have after transformation
    n_rows = sum(len(labelset) for labelset in y)

    if n_rows == X.shape[0]:
        # No expansion needed
        return X, flatten_list(y)

    # Performance improvements require csr matrix
    if not isinstance(X, csr_matrix):
        X = csr_matrix(X)

    indptr = np.zeros((n_rows+1), dtype=np.int32)
    indices = []
    data = []

    indices_count = 0
    offset = 0

    # Our goal is to expand the equal labelsets into their own row within X
    # We do this by repeating each row exactly "labelset" times
    for i, labelset in enumerate(y):
        labelset_sz = len(labelset)
        for j in range(labelset_sz):
            indptr[offset+j] = indices_count

            indices.append(X.indices[X.indptr[i]:X.indptr[i+1]])
            data.append(X.data[X.indptr[i]:X.indptr[i+1]])

            indices_count += len(X.data[X.indptr[i]:X.indptr[i+1]])

        offset += labelset_sz

    indptr[-1] = indices_count

    indices = np.concatenate(indices)
    data = np.concatenate(data)

    y_ = flatten_list(y)
    return csr_matrix((data, indices, indptr), shape=(n_rows, X.shape[1]), dtype=X.dtype), y_


def extract_rows_csr(matrix, rows):
    """
    Parameters
    ----------
    matrix : (sparse) csr_matrix
    rows : list of row ids
    Returns
    -------
    matrix_: (sparse) csr_matrix
        Transformed by extracting the desired rows from `matrix`
    """
    if not isinstance(matrix, csr_matrix):
        matrix = csr_matrix(matrix)

    # Short circuit if we want a blank matrix
    if len(rows) == 0:
        return csr_matrix(matrix.shape)

    # Keep a record of the desired rows
    indptr = np.zeros(matrix.indptr.shape, dtype=np.int32)
    indices = []
    data = []

    # Keep track of the current index pointer
    indices_count = 0

    for i in range(matrix.shape[0]):
        indptr[i] = indices_count

        if i in rows:
            indices.append(matrix.indices[matrix.indptr[i]:matrix.indptr[i+1]])
            data.append(matrix.data[matrix.indptr[i]:matrix.indptr[i+1]])
            indices_count += len(matrix.data[matrix.indptr[i]:matrix.indptr[i+1]])

    indptr[-1] = indices_count

    indices = np.concatenate(indices)
    data = np.concatenate(data)

    return csr_matrix((data, indices, indptr), shape=matrix.shape)


def nnz_rows_ix(X):
    """Return row indices which have at least one non-zero column value."""
    return np.unique(X.nonzero()[0])


def nnz_columns_count(X):
    """Return count of columns which have at least one non-zero value."""
    return len(np.count_nonzero(X, axis=0))


# + {"code_folding": []}
@logger
class HierarchicalClassifier(BaseEstimator, ClassifierMixin, MetaEstimatorMixin):
    """Hierarchical classification strategy
    Hierarchical classification deals with the scenario where our target classes have
    inherent structure that can be represented as a tree or a directed acyclic graph (DAG),
    with nodes representing the target classes themselves, and edges representing their inter-relatedness,
    e.g 'IS A' semantics.
    Within this general framework, several distinctions can be made based on a few key modelling decisions:
    - Multi-label classification - Do we support classifying into more than a single target class/label
    - Mandatory / Non-mandatory leaf node prediction - Do we require that classification always results with
      classes corresponding to leaf nodes, or can intermediate nodes also be treated as valid output predictions.
    - Local classifiers - the local (or "base") classifiers can theoretically be chosen to be of any kind, but we
      distinguish between three main modes of local classification:
            * "One classifier per parent node" - where each non-leaf node can be fitted with a multi-class
              classifier to predict which one of its child nodes is relevant for given example.
            * "One classifier per node" - where each node is fitted with a binary "membership" classifier which
              returns a binary (or a probability) score indicating the fitness for that node and the current
              example.
            * Global / "big bang" classifiers - where a single classifier predicts the full path in the hierarchy
              for a given example.
    The nomenclature used here is based on the framework outlined in [1].
    Parameters
    ----------
    base_estimator : classifier object, function, dict, or None
        A scikit-learn compatible classifier object implementing 'fit' and 'predict_proba' to be used as the
        base classifier.
        If a callable function is given, it will be called to evaluate which classifier to instantiate for
        current node. The function will be called with the current node and the graph instance.
        Alternatively, a dictionary mapping classes to classifier objects can be given. In this case,
        when building the classifier tree, the dictionary will be consulted and if a key is found matching
        a particular node, the base classifier pointed to in the dict will be used. Since this is most often
        useful for specifying classifiers on only a handlful of objects, a special 'DEFAULT' key can be used to
        set the base classifier to use as a catch all.
        If not provided, a base estimator will be chosen by the framework using various meta-learning
        heuristics (WIP).
    class_hierarchy : networkx.DiGraph object, or dict-of-dicts adjacency representation (see examples)
        A directed graph which represents the target classes and their relations. Must be a tree/DAG (no cycles).
        If not provided, this will be initialized during the `fit` operation into a trivial graph structure linking
        all classes given in `y` to an artificial "ROOT" node.
    prediction_depth : "mlnp", "nmlnp"
        Prediction depth requirements. This corresponds to whether we wish the classifier to always terminate at
        a leaf node (mandatory leaf-node prediction, "mlnp"), or wish to support early termination via some
        stopping criteria (non-mandatory leaf-node prediction, "nmlnp"). When "nmlnp" is specified, the
        stopping_criteria parameter is used to control the behaviour of the classifier.
    algorithm : "lcn", "lcpn"
        The algorithm type to use for building the hierarchical classification, according to the
        taxonomy defined in [1].
        "lcpn" (which is the default) stands for "local classifier per parent node". Under this model,
        a multi-class classifier is trained at each parent node, to distinguish between each child nodes.
        "lcn", which stands for "local classifier per node". Under this model, a binary classifier is trained
        at each node. Under this model, a further distinction is made based on how the training data set is constructed.
        This is controlled by the "training_strategy" parameter.
    training_strategy: "exclusive", "less_exclusive", "inclusive", "less_inclusive",
                       "siblings", "exclusive_siblings", or None.
        This parameter is used when the 'algorithm' parameter is to set to "lcn", and dictates how training data
        is constructed for training the binary classifier at each node.
    stopping_criteria: function, float, or None.
        This parameter is used when the 'prediction_depth' parameter is set to "nmlnp", and is used to evaluate
        at a given node whether classification should terminate or continue further down the hierarchy.
        When set to a float, the prediction will stop if the reported confidence at current classifier is below
        the provided value.
        When set to a function, the callback function will be called with the current node attributes,
        including its metafeatures, and the current classification results.
        This allows the user to define arbitrary logic that can decide whether classification should stop at
        the current node or continue. The function should return True if classification should continue,
        or False if classification should stop at current node.
    root : integer, string
        The unique identifier for the qualified root node in the class hierarchy. The hierarchical classifier
        assumes that the given class hierarchy graph is a rooted DAG, e.g has a single designated root node
        of in-degree 0. This node is associated with a special identifier which defaults to a framework provided one,
        but can be overridden by user in some cases, e.g if the original taxonomy is already rooted and there's no need
        for injecting an artifical root node.
    progress_wrapper : progress generator or None
        If value is set, will attempt to use the given generator to display progress updates. This added functionality
        is especially useful within interactive environments (e.g in a testing harness or a Jupyter notebook). Setting
        this value will also enable verbose logging. Common values in tqdm are `tqdm_notebook` or `tqdm`
    Attributes
    ----------
    classes_ : array, shape = [`n_classes`]
        Flat array of class labels
    References
    ----------
    .. [1] CN Silla et al., "A survey of hierarchical classification across
           different application domains", 2011.
    """

    def __init__(self, base_estimator=None, class_hierarchy=None, prediction_depth="mlnp",
                 algorithm="lcpn", training_strategy=None, stopping_criteria=None,
                 root=None, progress_wrapper=None):
        self.base_estimator = base_estimator
        self.class_hierarchy = class_hierarchy
        self.prediction_depth = prediction_depth
        self.algorithm = algorithm
        self.training_strategy = training_strategy
        self.stopping_criteria = stopping_criteria
        self.root = root
        self.progress_wrapper = progress_wrapper
        #print(root)

    def fit(self, X, y=None, sample_weight=None):
        """Fit underlying classifiers.
        Parameters
        ----------
        X : (sparse) array-like, shape = [n_samples, n_features]
            Data.
        y : (sparse) array-like, shape = [n_samples, ], [n_samples, n_classes]
            Multi-class targets. An indicator matrix turns on multilabel
            classification.
        sample_weight : array-like, shape (n_samples,), optional (default=None)
            Weights applied to individual samples (1. for unweighted).
        Returns
        -------
        self
    
        """
        
        X, y = check_X_y(X, y, accept_sparse='csr')
        check_classification_targets(y)
        if sample_weight is not None:
            check_consistent_length(y, sample_weight)

        # Check that parameter assignment is consistent
        self._check_parameters()

        # Initialize NetworkX Graph from input class hierarchy
        self.class_hierarchy_ = self.class_hierarchy or make_flat_hierarchy(list(np.unique(y)), root=self.root)
        self.graph_ = DiGraph(self.class_hierarchy_)
        self.is_tree_ = is_tree(self.graph_)
        self.classes_ = list(
            node
            for node in self.graph_.nodes()
            if node != self.root
        )
        self.logger.debug("fit() - self.classes_ = %s", self.classes_)

        # Recursively build training feature sets for each node in graph
        with self._progress(total=self.n_classes_ + 1, desc="Building features") as progress:
            self._recursive_build_features(X, y, node_id=self.root, progress=progress)

        # Recursively train base classifiers
        with self._progress(total=self.n_classes_ + 1, desc="Training base classifiers") as progress:
            self._recursive_train_local_classifiers(X, y, node_id=self.root, progress=progress)

        return self

    def predict(self, X):
        """Predict multi-class targets using underlying estimators.
        Parameters
        ----------
        X : (sparse) array-like, shape = [n_samples, n_features]
            Data.
        Returns
        -------
        y : (sparse) array-like, shape = [n_samples, ], [n_samples, n_classes].
            Predicted multi-class targets.
        """
        check_is_fitted(self, "graph_")
        X = check_array(X, accept_sparse="csr")

        def _classify(x, progress):
            #with self._progress(total=self.n_classes_ + 1, desc="Predicting") as progress:
            path, _ = self._recursive_predict(x, root=self.root, progress=progress)
            return path[-1]
        
        with self._progress(total=X.shape[0], desc="Predicting") as progress:
            y_pred = apply_along_rows( self, _classify, progress, X=X)
        return y_pred

    def predict_proba(self, X):
        """
        Return probability estimates for the test vector X.
        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
        Returns
        -------
        C : array-like, shape = [n_samples, n_classes]
            Returns the probability of the samples for each class in
            the model. The columns correspond to the classes in sorted
            order, as they appear in the attribute `classes_`.
        """
        check_is_fitted(self, "graph_")
        X = check_array(X, accept_sparse="csr")

        def _classify(x):
            _, scores = self._recursive_predict(x, root=self.root)
            return scores

        y_pred = apply_along_rows(_classify, X=X)
        return y_pred

    @property
    def n_classes_(self):
        return len(self.classes_)

    def _check_parameters(self):
        """Check the parameter assignment is valid and internally consistent."""
        validate_parameters(self)

    def _recursive_build_features(self, X, y, node_id, progress):
        """
        Build the training feature matrix X recursively, for each node.
        By default we use "hierarchical feature set" (terminology per Ceci and Malerba 2007)
        which builds up features at each node in the hiearchy by "rolling up" training examples
        defined on the the leaf nodes (classes) of the hierarchy into the parent classes relevant
        for classification at a particular non-leaf node.
        """
        #print('HOOO')
        if "X" in self.graph_.node[node_id]:
            #print('HO', node_id)
            # Already visited this node in feature building phase
            return self.graph_.node[node_id]["X"]

        self.logger.debug("Building features for node: %s", node_id)
        progress.update(1)

        if self.graph_.out_degree(node_id) == 0:
            # Leaf node
            indices = np.flatnonzero(y == node_id)
            self.graph_.node[node_id]["X"] = self._build_features(
                X=X,
                y=y,
                indices=indices,
            )
            return self.graph_.node[node_id]["X"]

        # Non-leaf node
        self.graph_.node[node_id]["X"] = csr_matrix(
            X.shape,
            dtype=X.dtype,
        )
        for child_node_id in self.graph_.successors(node_id):
            self.graph_.node[node_id]["X"] += \
                self._recursive_build_features(
                    X=X,
                    y=y,
                    node_id=child_node_id,
                    progress=progress,
            )

        # Build and store meta-features for node
        self.graph_.node[node_id][METAFEATURES] = self._build_metafeatures(
            X=self.graph_.node[node_id]["X"],
            y=y,
        )

        return self.graph_.node[node_id]["X"]

    def _build_features(self, X, y, indices):
        X_ = extract_rows_csr(X, indices)

        # Perform feature selection
        X_ = self._select_features(X=X_, y=np.array(y)[indices])

        return X_

    def _select_features(self, X, y):
        """
        Perform feature selection for training data.
        Can be overridden by a sub-class to implement feature selection logic.
        """
        return X

    def _build_metafeatures(self, X, y):
        """
        Build the meta-features associated with a particular node.
        These are various features that can be used in training and prediction time,
        e.g the number of training samples available for the classifier trained at that node,
        the number of targets (classes) to be predicted at that node, etc.
        Parameters
        ----------
        X : (sparse) array-like, shape = [n_samples, n_features]
            The training data matrix at current node.
        Returns
        -------
        metafeatures : dict
            Python dictionary of meta-features. The following meta-features are computed by default:
            * 'n_samples' - Number of samples used to train classifier at given node.
            * 'n_targets' - Number of targets (classes) to classify into at given node.
        """
        # Indices of non-zero rows in X, i.e rows corresponding to relevant samples for this node.
        ix = nnz_rows_ix(X)

        return dict(
            n_samples=len(ix),
            n_targets=len(np.unique(y[ix])),
        )

    def _recursive_train_local_classifiers(self, X, y, node_id, progress):
        if CLASSIFIER in self.graph_.node[node_id]:
            # Already trained classifier at this node, skip
            return

        progress.update(1)
        self._train_local_classifier(X, y, node_id)

        for child_node_id in self.graph_.successors(node_id):
            self._recursive_train_local_classifiers(
                X=X,
                y=y,
                node_id=child_node_id,
                progress=progress,
            )

    def _train_local_classifier(self, X, y, node_id):
        if self.graph_.out_degree(node_id) == 0:
            # Leaf node
            if self.algorithm == "lcpn":
                #print('leaf')
                # Leaf nodes do not get a classifier assigned in LCPN algorithm mode.
                self.logger.debug(
                    "_train_local_classifier() - skipping leaf node %s when algorithm is 'lcpn'",
                    node_id,
                )
                return
            
        #print('here')

        X = self.graph_.node[node_id]["X"]
        nnz_rows = nnz_rows_ix(X)
        X_ = X[nnz_rows, :]

        y_rolled_up = rollup_nodes(
            graph=self.graph_,
            source=node_id,
            targets=[y[idx] for idx in nnz_rows],
        )

        if self.is_tree_:
            #print('tree')
            y_ = flatten_list(y_rolled_up)
        else:
            # Class hierarchy graph is a DAG
            X_, y_ = apply_rollup_Xy(X_, y_rolled_up)

        num_targets = len(np.unique(y_))

        self.logger.debug(
            "_train_local_classifier() - Training local classifier for node: %s, X_.shape: %s, len(y): %s, n_targets: %s",  # noqa:E501
            node_id,
            X_.shape,
            len(y_),
            num_targets,
        )
       # print(
       #     "_train_local_classifier() - Training local classifier for node: %s, X_.shape: %s, len(y): %s, n_targets: %s, targets: %s" %  # noqa:E501
       #     (node_id,
       #     X_.shape,
       #     len(y_),
       #     num_targets,
       #     np.unique(y_))
       # )


        if X_.shape[0] == 0:
            # No training data could be materialized for current node
            #print('no training data')
            self.logger.warning(
                "_train_local_classifier() - not enough training data available to train, classification in branch will terminate at node %s",  # noqa:E501
                node_id,
            )
            return
        elif num_targets == 1:
            # Training data could be materialized for only a single target at current node
            #print('only single target')
            constant = y_[0]
            self.logger.debug(
                "_train_local_classifier() - only a single target (child node) available to train classifier for node %s, Will trivially predict %s",  # noqa:E501
                node_id,
                constant,
            )

            clf = DummyClassifier(strategy="constant", constant=constant)
        else:
            clf = self._base_estimator_for(node_id)
        
        #print('fit', node_id)
        clf.fit(X=X_, y=y_)
        self.graph_.node[node_id][CLASSIFIER] = clf

    def _recursive_predict(self, x, root, progress):
        clf = self.graph_.node[root][CLASSIFIER]
        path = [root]
        path_proba = []
        class_proba = np.zeros_like(self.classes_, dtype=np.float64)
        progress.update(1)
        
        while clf:
            probs = clf.predict_proba(x)[0]
            argmax = np.argmax(probs)
            score = probs[argmax]
            path_proba.append(score)
            
           

            # Report probabilities in terms of complete class hierarchy
            for local_class_idx, class_ in enumerate(clf.classes_):
                try:
                    class_idx = self.classes_.index(class_)
                except ValueError:
                    # This may happen if the classes_ enumeration we construct during fit()
                    # has a mismatch with the individual node classifiers' classes_.
                    self.logger.error(
                        "Could not find index in self.classes_ for class_ = '%s' (type: %s). path: %s",
                        class_,
                        type(class_),
                        path,
                    )
                    raise

                class_proba[class_idx] = probs[local_class_idx]
                if local_class_idx == argmax:
                    prediction = class_

            if self._should_early_terminate(
                current_node=path[-1],
                prediction=prediction,
                score=score,
            ):
                break

            # Update current path
            path.append(prediction)

            clf = self.graph_.node[prediction].get(CLASSIFIER, None)

        return path, class_proba

    def _should_early_terminate(self, current_node, prediction, score):
        """
        Evaluate whether classification should terminate at given step.
        This depends on whether early-termination, as dictated by the the 'prediction_depth'
          and 'stopping_criteria' parameters, is triggered.
        """
        if self.prediction_depth != "nmlnp":
            # Prediction depth parameter does not allow for early termination
            return False

        if (
            isinstance(self.stopping_criteria, float)
            and score < self.stopping_criteria
        ):
            if current_node == self.root:
                return False

            self.logger.debug(
                "_should_early_terminate() - score %s < %s, terminating at node %s",
                score,
                self.stopping_criteria,
                current_node,
            )
            return True
        elif callable(self.stopping_criteria):
            return self.stopping_criteria(
                current_node=self.graph_.nodes[current_node],
                prediction=prediction,
                score=score,
            )

        return False

    def _base_estimator_for(self, node_id):
        base_estimator = None
        if self.base_estimator is None:
            # No base estimator specified by user, try to pick best one
            base_estimator = self._make_base_estimator(node_id)

        elif isinstance(self.base_estimator, dict):
            # User provided dictionary mapping nodes to estimators
            if node_id in self.base_estimator:
                base_estimator = self.base_estimator[node_id]
            else:
                base_estimator = self.base_estimator[DEFAULT]

        elif is_estimator(self.base_estimator):
            # Single base estimator object, return a copy
            base_estimator = self.base_estimator

        else:
            # By default, treat as callable factory
            base_estimator = self.base_estimator(node_id=node_id, graph=self.graph_)

        return clone(base_estimator)

    def _make_base_estimator(self, node_id):
        return LogisticRegression()

    def _progress(self, total, desc, **kwargs):
        if self.progress_wrapper:
            return self.progress_wrapper(total=total, desc=desc)
        else:
            return DummyProgress()


# -

def lcpn_classic(root, tree, classifier_object, X, y, testing_data):

    ROOT = root
    #print(root)

    class_hierarchy = tree
    
    base_estimator = Pipeline([
    #('vect', CountVectorizer()),
    #('tfidf', TfidfTransformer()),
    ('clf', classifier_object),
])
        
    clf = HierarchicalClassifier(
        base_estimator=base_estimator,
        class_hierarchy=class_hierarchy, root=root, progress_wrapper=tqdm_notebook)



    #clf.fit(X_train, y_train)
    vectorizer = TfidfVectorizer()
    clf.fit(vectorizer.fit_transform(X), y)
    y_pred = clf.predict(vectorizer.transform(testing_data))
    
    return y_pred

# +
#try out lcpn classic
y_pred = lcpn_classic('Computers & Electronics', tree_icecat, MultinomialNB(), training_data['Title'].values, training_target['Category.Name.Value'].values, testing_data['Title'].values)



# -

h_f_score = hierarchical_eval(testing_target['Category.Name.Value'].values, y_pred, tree_icecat, 'Computers & Electronics')


# ## LCPN advanced

# +
def train_lcpn_advanced(root, tree, pipeline, X, y):
    pipelines_dict_tmp={} 
    training_column = X.columns.tolist()[0]
    target_column = y.columns.tolist()[0]
    warnings.filterwarnings("ignore") 
    
    #build tmp data. No cv, as Steiner (oder so?) found there is no use and computationlly expensive
    data = pd.concat([X, y], axis=1)
    data = data.loc[:,~data.columns.duplicated()]
    
    with tqdm_notebook(total=len(tree.nodes), desc="Training classifiers") as progress:
        for node in tree.nodes:
            progress.update(1)
            
            successors_direct = set(tree.successors(node))
            successors_all_dict = nx.dfs_successors(tree, source=node)
            successors_all = set()
            for key, value in successors_all_dict.items():
                successors_all.update(value)

            #leaf node -> no classifier needed
            if len(successors_all) == 0:
                continue

            tmp = data[data[target_column].isin(successors_all)]
            tmp_training_data = tmp[[training_column]]
            tmp_training_target = tmp[[target_column]]

            #no examples for that node in dataset
            if len(tmp_training_target) == 0:
                continue

            #fit classifier for node
            pipelines_dict_tmp[node] = copy.deepcopy(pipeline)
            #Only allowed to have direct successors as targets -> relabel targets
            for index, row in tmp_training_target.iterrows():
                if row[target_column] in successors_direct:
                    continue
                parent = list(tree.predecessors(row[target_column]))[0]
                while parent not in successors_direct:
                    parent = list(tree.predecessors(parent))[0]

                tmp_training_target.at[index,target_column] = parent

            pipelines_dict_tmp[node].fit(tmp_training_data[training_column].values.astype('U'), tmp_training_target[target_column].values)
            
    return pipelines_dict_tmp
    
#apply classifiers
def apply_lcpn_advanced(root, pipelines_dict, testing_data):
    training_column = testing_data.columns.tolist()[0]
    with tqdm_notebook(total=len(testing_data), desc="Applying classifiers") as progress:
        y_pred=[]
        for index, row in testing_data.iterrows():
            progress.update(1)
            #make prediction at root
            y_pred_tmp = pipelines_dict[root].predict([row[training_column]])
            while y_pred_tmp[0] in pipelines_dict.keys():
                y_pred_tmp = pipelines_dict[y_pred_tmp[0]].predict([row[training_column]])
            y_pred.append(y_pred_tmp[0])
        return y_pred
            
            
#####################
def apply_lcpn_advanced_parallel(root, pipelines_dict, row):
    
    y_pred_tmp = pipelines_dict[root].predict([row])
    while y_pred_tmp[0] in pipelines_dict.keys():
        y_pred_tmp = pipelines_dict[y_pred_tmp[0]].predict([row])
        
    return y_pred_tmp                                                                     
            
        

# +
# compare the other implementation that I adapted to make it work for me to mine on the same small sample set as well
pipeline = Pipeline([
    ('vect', CountVectorizer()),
    ('tfidf', TfidfTransformer()),
    ('clf', MultinomialNB()),
])

pipelines_dict = train_lcpn_advanced('Computers & Electronics', tree_icecat, pipeline, training_data, training_target)



# -

y_pred = apply_lcpn_advanced('Computers & Electronics', pipelines_dict, testing_data)

h_f_score = hierarchical_eval(testing_target['Category.Name.Value'].values, y_pred, tree_icecat, 'Computers & Electronics')


# -> My score is 2% points higher than the other implementation that I changed to make it work for me -> use my implementation from here on

# +
# apply classifiers parrralel
with PoolExecutor(max_workers=10) as executor:
    futures_pred = list(tqdm_notebook(executor.map(apply_lcpn_advanced_parallel, repeat('Computers & Electronics'), repeat(pipelines_dict), testing_data['Title'].values),total=len(testing_data[['Title']].index.values), desc="Applying classifiers"))


# -

# -> Applying in parrallel does not give significant speed increase locally, but on Server 2x as fast

flat_list = [item for sublist in futures_pred for item in sublist]

h_f_score = hierarchical_eval(testing_target['Category.Name.Value'].values, flat_list, tree_icecat, 'Computers & Electronics')


# ## LCPN Icecat

# +
pipeline = Pipeline([
    ('vect', CountVectorizer()),
    ('tfidf', TfidfTransformer()),
    ('clf', MultinomialNB()),
])

pipelines_dict = train_lcpn_advanced('Computers & Electronics', tree_icecat, pipeline, icecat_data_train[['Title']], icecat_data_train_target[['Category.Name.Value']])


# +
pipeline = Pipeline([
    ('vect', CountVectorizer()),
    ('tfidf', TfidfTransformer()),
    ('clf', MultinomialNB()),
])

pipelines_dict = train_lcpn_advanced('Computers & Electronics', tree_icecat, pipeline, icecat_data_train[['Title']], icecat_data_train_target[['Category.Name.Value']])

# -

y_pred = apply_lcpn_advanced('Computers & Electronics', pipelines_dict, icecat_data_test[['Title']])
y_true = icecat_data_test['Category.Name.Value'].values

# +
# apply classifiers parrralel
#with PoolExecutor(max_workers=10) as executor:
#    futures_pred = list(tqdm_notebook(executor.map(apply_lcpn_advanced_parallel, repeat('Computers & Electronics'), repeat(pipelines_dict), icecat_data_test['Title'].values),total=len(icecat_data_test[['Title']].index.values), desc="Applying classifiers"))


# -

# Traditional testing
scores_all_labels_false, scores_all_labels_true = eval_traditional(e_2, y_true, y_pred)

# Hierarchical testing
h_f_score = hierarchical_eval(e_2, y_true, y_pred, tree_icecat, 'Computers & Electronics')


# -> Beat flat classification

# ## LCPN WDC

# use pipelines_dict trained on icecat
y_pred = apply_lcpn_advanced('Computers & Electronics', pipelines_dict, wdc_data_full[['title']])
y_true = wdc_data_full['CategoryName'].values

# +
# Traditional testing
scores_all_labels_false, scores_all_labels_true = eval_traditional(e_2, y_true, y_pred)


# -

# Hierarchical testing
h_f_score = hierarchical_eval(e_2, y_true, y_pred, tree_wdc, 'Computers & Electronics')

y_pred = apply_lcpn_advanced('Computers & Electronics', pipelines_dict, (wdc_data_full['title'].str.strip() + ' ' + wdc_data_full['desc'].str.strip()).to_frame())


# Traditional testing
scores_all_labels_false, scores_all_labels_true = eval_traditional(e_2_wdc, y_true, y_pred)

# Hierarchical testing
h_f_score = hierarchical_eval(e_2_wdc, y_true, y_pred, tree_wdc, 'Computers & Electronics')


# ## LCPN Rakuten

rakuten_data_train.head()

len(rakuten_data_train)

# +
pipeline = Pipeline([
    ('vect', CountVectorizer()),
    ('tfidf', TfidfTransformer()),
    ('clf', MultinomialNB()),
])

pipelines_dict = train_lcpn_advanced('Root', tree_rakuten, pipeline, rakuten_data_train[['title']], rakuten_data_train[['CategoryID']])

# -

y_pred = apply_lcpn_advanced('Root', pipelines_dict, rakuten_data_test[['title']])

with open('y_pred_rakuten.pkl', 'wb') as f:
    pickle.dump(y_pred, f)

y_true = rakuten_data_test['CategoryID'].values

# Traditional testing
scores_all_labels_false, scores_all_labels_true = eval_traditional(e_2, y_true, y_pred)

# Hierarchical testing
h_f_score = hierarchical_eval(e_2, y_true, y_pred, tree_rakuten, 'Root')


# # Fasttext

# prepare data
def prepare_fasttext(X, y=None):
    if y is None:
        fasttext_full = copy.deepcopy(X)
        fasttext_full[fasttext_full.columns[0]] = fasttext_full[fasttext_full.columns[0]].str.replace(' ', '_')
        fasttext_full[fasttext_full.columns[0]]  = '__label__' + fasttext_full[fasttext_full.columns[0]].astype(str)
        return fasttext_full[fasttext_full.columns[0]].values
    
    fasttext_full = pd.concat([X, y], axis=1)
    
    cols = fasttext_full.columns.tolist()
    cols = cols[-1:] + cols[:-1]
    fasttext_full = fasttext_full[cols]
    
    fasttext_full[fasttext_full.columns[0]] = fasttext_full[fasttext_full.columns[0]].str.replace(' ', '_')
    fasttext_full[fasttext_full.columns[0]]  = '__label__' + fasttext_full[fasttext_full.columns[0]].astype(str)
    #fasttext_full['Title'] = fasttext_full['Title'].str.replace('"', '')

    fasttext_full.to_csv(dataPathSets + 'fasttext_train' + '.txt', index=False, sep=' ', header=False, quoting=csv.QUOTE_NONE, quotechar="", escapechar=" ")
    
    return fasttext_full
    


# + {"code_folding": []}
# fasttext with default parameters except loss='ns' because Rakuten used this as well
def train_fasttext(X_train, y_train, epoch=5, lr=0.1):
    #prepare data
    fasttext_train = prepare_fasttext(X_train, y_train)
    
    #train model
    model = fasttext.train_supervised(input=dataPathSets+"fasttext_train.txt", epoch=epoch, lr=lr, loss='ns')
    
    return model
    


# -

def apply_fasttext(model, X_test):
    #prepare data
    #fasttext_test = prepare_fasttext(X_test, y_test,'test')
    
    #apply model
    list_to_predict = list(X_test[X_test.columns[0]].values)
    list_to_predict = [item.replace('\n', ' ').replace('\r', '') for item in list_to_predict]
    y_pred = model.predict(list_to_predict)[0]
    y_pred = [item for sublist in y_pred for item in sublist]
    
    return y_pred
    


# ## Fasttext Icecat sample

model = train_fasttext(training_data,training_target)

y_pred = apply_fasttext(model, testing_data)

y_true = prepare_fasttext(testing_target)

# Traditional testing
scores_all_labels_false, scores_all_labels_true = eval_traditional('Fasttext', y_true, y_pred)

#tune fasttext
model = train_fasttext(training_data,training_target, epoch=25, lr=1.0)

y_pred = apply_fasttext(model, testing_data)

# Traditional testing
scores_all_labels_false, scores_all_labels_true = eval_traditional('Fasttext', y_true, y_pred)

# Bring labels into form of node names
y_pred = np.asarray([i.replace('__label__', '').replace('_', ' ') for i in y_pred]) 
y_true = np.asarray([i.replace('__label__', '').replace('_', ' ') for i in y_true])

# Hierarchical testing
h_f_score = hierarchical_eval(y_true, y_pred, tree_icecat, 'Computers & Electronics')

# ## Fasttext Icecat

model = train_fasttext(icecat_data_train[['Title']], icecat_data_train_target[['Category.Name.Value']]) 

y_pred = apply_fasttext(model, icecat_data_test[['Title']])

y_true = prepare_fasttext(icecat_data_test[['Category.Name.Value']])

# Traditional testing
scores_all_labels_false, scores_all_labels_true = eval_traditional(e_3, y_true, y_pred)

#tune fasttext
model = train_fasttext(icecat_data_train[['Title']], icecat_data_train_target[['Category.Name.Value']], epoch=25, lr=1.0) 

y_pred = apply_fasttext(model, icecat_data_test[['Title']])

# Traditional testing
scores_all_labels_false, scores_all_labels_true = eval_traditional(e_3, y_true, y_pred)

# +
# Hierarchical testing
# Bring labels into form of node names
y_pred = np.asarray([i.replace('__label__', '').replace('_', ' ') for i in y_pred]) 
y_true = np.asarray([i.replace('__label__', '').replace('_', ' ') for i in y_true])

h_f_score = hierarchical_eval(e_3, y_true, y_pred, tree_icecat, 'Computers & Electronics')
# -

# ## Fasttext WDC

# use model from icecat
y_pred = apply_fasttext(model, wdc_data_full[['title']])

y_true = prepare_fasttext(wdc_data_full[['CategoryName']])

# Traditional testing
scores_all_labels_false, scores_all_labels_true = eval_traditional(e_3, y_true, y_pred)

# +
# Hierarchical testing
# Bring labels into form of node names
y_pred = np.asarray([i.replace('__label__', '').replace('_', ' ') for i in y_pred]) 
y_true = np.asarray([i.replace('__label__', '').replace('_', ' ') for i in y_true])

h_f_score = hierarchical_eval(e_3, y_true, y_pred, tree_wdc, 'Computers & Electronics')
# -

y_pred = apply_fasttext(model, (wdc_data_full['title'].str.strip() + ' ' + wdc_data_full['desc'].str.strip()).to_frame())
y_true = prepare_fasttext(wdc_data_full[['CategoryName']])

# Traditional testing
scores_all_labels_false, scores_all_labels_true = eval_traditional(e_3_wdc, y_true, y_pred)

# +
# Hierarchical testing
y_pred = np.asarray([i.replace('__label__', '').replace('_', ' ') for i in y_pred]) 
y_true = np.asarray([i.replace('__label__', '').replace('_', ' ') for i in y_true])

h_f_score = hierarchical_eval(e_3, y_true, y_pred, tree_wdc, 'Computers & Electronics')
# -

# ## Fasttext Rakuten

model = train_fasttext(rakuten_data_train[['title']], rakuten_data_train[['CategoryID']], epoch=25, lr=1.0) 

y_pred = apply_fasttext(model, rakuten_data_test[['title']])

y_true = prepare_fasttext(rakuten_data_test[['CategoryID']])

# Traditional testing
scores_all_labels_false, scores_all_labels_true = eval_traditional(e_3, y_true, y_pred)

# +
# Hierarchical testing
# Bring labels into form of node names
y_pred = np.asarray([i.replace('__label__', '').replace('_', ' ') for i in y_pred]) 
y_true = np.asarray([i.replace('__label__', '').replace('_', ' ') for i in y_true])

h_f_score = hierarchical_eval(e_3, y_true, y_pred, tree_rakuten, 'Root')
# -

# # Random

training_data, testing_data, training_target, testing_target = create_samples(icecat_data_full, ['Title', 'Description.LongDesc'], ['Category.Name.Value'], 0.01, 0.01)

len(testing_target)

# +
# Traditional testing
# Evaluate pipeline from cv on held out testing set. Refit on whole training set first

# Traditional cv
# Naive Bayes as competitive baseline
# Use pipeline: here, different models and model-parameters can be put in and tested using CV
pipeline = Pipeline([
    ('vect', CountVectorizer()),
    ('tfidf', TfidfTransformer()),
    ('clf', MultinomialNB()),
])
classifier = pipeline.fit(training_data['Title'].values, training_target['Category.Name.Value'].values)




# +
# Traditional testing
# Apply classifier trained on Icecat training set
# classifier.predict transforms data according to pipeline (classifier is a pipeline object) 
# and predicts with the final estimator

y_pred = classifier.predict(wdc_data_full_testing['title'].values)
y_true = wdc_data_full_testing['CategoryName'].values
# -

p = classifier.predict(testing_data['Title'].values)

testing_target['Category.Name.Value'].values

hierarchical_eval(testing_target['Category.Name.Value'].values, p, tree_icecat, 'Computers & Electronics')

a = [[2,3],[4,5]]

a[:,0]

b = np.asarray(a)

type(a)

b[:,1]

b

wdc_only_titles = wdc_data_full[wdc_data_full['title']!='']

len(wdc_only_titles['CategoryName'].unique())

len(wdc_data_full['CategoryName'].unique())

wdc_data_full.columns

icecat_data_full.columns

ç[['GTIN']].tolist().merge(wdc_data_full[['gtin']].tolist(), left_on='GTIN', right_on='gtin', how='left')

icecat_data_join=copy.deepcopy(icecat_data_full)

wdc_join = copy.deepcopy(wdc_data_full)

for index, row in wdc_join.iterrows():
    wdc_join.at[index,['gtin']] = ', '.join(row['gtin'])

gtins_wdc=list(wdc_data_full['gtin'])



wdc_data_full.head(100).sort_values('title')

icecat_join_empty = icecat_data_join.iloc[0:0]


wdc_join_empty

with tqdm_notebook(total=len(icecat_data_full)) as progress:
    for index, row in icecat_data_full.iterrows():
        progress.update(1)

        if len(row['GTIN'])>0:
            if row['GTIN'][0] in gtins_wdc:
                #print(row)
                icecat_join_empty=icecat_join_empty.append(row)
    

icecat_join_empty[['GTIN','Title', 'Category.Name.Value', 'Description.LongDesc', 'SummaryDescription.LongSummaryDescription']].sort_values('GTIN')

wdc_data_full[['title','desc','CategoryName','gtin']].sort_values('gtin')[100:]


pd.merge(icecat_data_join, wdc_join, how='left', left_on=['GTIN'], right_on=['gtin'])

tree_icecat.nodes['Printing Equipment'].

import random
#random.sample(range(1, 700000), 100)
for i in random.sample(range(1, 700000), 100):
    print(icecat_data_full.iloc[i]['pathlist_names'])
#icecat_data_full[['pathlist_names']]

for n in tree_icecat.successors('Printing Supplies'):
    print(n)

with tqdm_notebook(total=len(icecat_data_full)) as progress:
    
    for index, row in icecat_data_full.iterrows():
        progress.update(1)
        if len(row['pathlist_names'].split('>'))>4:
            print(row['pathlist_names'])

nx.shortest_path_length(tree_icecat, source='Computers & Electronics', target='Ink Sticks')

for n in tree_rakuten.successors('4559'):
    print(n)

import random
#random.sample(range(1, 700000), 100)
for i in random.sample(range(1, 800000), 700000):
    if len(rakuten_data_full.iloc[i]['path'].split('>')) > 7:
        print(rakuten_data_full.iloc[i]['path'])
#icecat_data_full[['pathlist_names']]

# ## 


